/*-
 * definitions of opcode analyzer and break table header.
 *
 */

#ifndef analyze_h
#define analyze_h

#ifndef __TYPES_H_INCLUDED
#include <sys/types.h>
#endif
#ifndef _LIMITS_H_INCLUDED
#include <limits.h>
#endif

#define MAXFOL 2
typedef struct {
	char	*name;
	uchar_t	flags;
	uchar_t	nfol;
#define _OP_ISJMP	1
	struct {
		uchar_t	flags;
#define _OPR_ISREL	1
		unsigned offset;
	} foll[MAXFOL];
} OpCode;



typedef struct {
	ulong_t   magic;
	char      image[PATH_MAX+1];
#define BMAGIC 01732145
	ulong_t   bsize;
	ulong_t   base;
} BreakHeader;


typedef struct {
	BreakHeader      bh;
	uchar_t         *bc;
} BreakTab;

enum {
	BranchTarget      = 1<<0,
	Terminal          = 1<<1,
	Visited           = 1<<2,

	UserBit           = 1<<7
};
#define BrMask     (BranchTarget|Terminal|Visited)

#define br_set(b,n,m)     ((b)->bc[(n)]  |=  (m))
#define br_clr(b,n,m)     ((b)->bc[(n)]  &= ~(m))
#define br_isset(b,n,m)  (((b)->bc[(n)]  &   (m)) != 0)

#define br_mark(b,n)      br_set(b,n, BranchTarget)
#define br_unmark(b,n)    br_clr(b,n, BranchTarget)
#define br_ismarked(b,n)  br_isset(b,n,BranchTarget)

#define br_term(b,n)      br_set(b,n, Terminal)
#define br_nonterm(b,n)   br_clr(b,n, Terminal)
#define br_isterm(b,n)    br_isset(b,n, Terminal)

#define br_visit(b,n)     br_set(b, n, Visited)
#define br_novisit(b,n)   br_clr(b, n, Visited)
#define br_isvisit(b,n)   br_isset(b,n, Visited)

#define br_user(b,n)     br_set(b, n, UserBit)
#define br_nouser(b,n)   br_clr(b, n, UserBit)
#define br_isuser(b,n)   br_isset(b,n, UserBit)



extern int anaylze_opcode(uchar_t *op, OpCode *oc);

extern int braflow(uchar_t *text, BreakTab *b, ulong_t start,
                   ulong_t lower, ulong_t upper);

extern BreakTab *get_brtab(char *fname);
extern int       put_brtab(char *fname, BreakTab *b);

extern BreakTab *new_brtab(char *src, ulong_t base, ulong_t siz);
extern int       delete_brtab(BreakTab *b);

#endif
