
#include <libc.h>
#include "analyze.h"
#include "dbi.h"

int verbose;




print_info(lmfo *lo, ulong_t a0, ulong_t a1)
{
	char         buf0[256], buf1[256];
	unsigned     lno0, lno1;
	buf0[0] = '\0';
	buf1[0] = '\0';

	if (lo == 0) {
		if (verbose)
			printf("0x%lx..0x%lx\n",a0, a1);
		return 0;
	}

	if (lmf_a2l(lo, a0, buf0, &lno0)==-1 ||
	    lmf_a2l(lo, a1, buf1, &lno1)==-1) {
	    	if (verbose)
	    		printf("(nosyms) 0x%lx .. 0x%lx\n", a0, a1);
	    	return 0;
	}
	if (lno0 != -1) {
		if (lno1 != -1) {
			if (strcmp(buf0, buf1) == 0) {
				printf("%s line %u..%u\n", buf0, lno0, lno1);
			} else {
				printf("%s line %u-\n", buf0, lno0);
				printf("%s line - %u\n", buf1, lno1);
			}
		} else {
			if (lmf_a2m(lo, a1, buf1, &lno1) == -1) {
				fprintf(stderr, "panic: unreachable module\n");
				exit(1);
			}
			printf("%s line %u - module %s offset 0x%lx\n",
				buf0, lno0, buf1, lno1);
		}
	} else {
		if (lmf_a2m(lo, a0, buf0, &lno0) == -1) {
			fprintf(stderr, "panic: unreachable module\n");
			exit(1);
		}
		if (lno1 != -1) {
			printf("module %s offset 0x%x - %s + %u\n",
				buf0, lno0, buf1, lno1);
		} else if (verbose) {
			if (lmf_a2m(lo, a1, buf1, &lno1) == -1) {
				fprintf(stderr, "panic: unreachable module\n");
				exit(1);
			}
			if (strcmp(buf0, buf1) == 0) {
				printf("module %s offset 0x%x - 0x%x\n",
					buf0, lno0, lno1);
			} else {
				printf("module %s offset 0x%x - module %s offset 0x%x\n",
					buf0, lno0, buf1, lno1);
			}
		} 
	}
	return 0;
}

brdiff(BreakTab *src, BreakTab *res)
{
	unsigned n;
	unsigned save_n;
	unsigned state = 0;
	lmfo    *lo;
	FILE    *fp;
	if ((fp=fopen(src->bh.image, "r")) == 0 ||
	    (lo=lmf_open(fp)) == 0) {
		fprintf(stderr, "warning, unable to open '%s'\n",
			src->bh.image);
	}
	for (n=0; n < src->bh.bsize; n++) {
		switch (state) {
		case 0:
			if (br_ismarked(src, n) && br_ismarked(res, n)) {
				save_n = n;
				state = 1;
			}
			break;
		case 1:
			if (br_ismarked(src, n)) {
				if (br_ismarked(res, n)) {
					print_info(lo, src->bh.base+save_n,
					               src->bh.base+n-1);
					state = 0;
				}
			}
			break;
		}
	}
	if (state == 1) {
		print_info(lo, src->bh.base+save_n,
		               src->bh.base+n-1);
	}
}

main(int argc, char **argv)
{
	int	c,i;
	BreakTab *orig, *output;
	FILE     *f;
	while ((c=getopt(argc, argv, "v")) != -1) {
		switch (c) {
		case 'v': verbose = 1; break;
		default:
			break;
		}
	}
	if (optind != argc-2) {
		fprintf(stderr, "usage %s [options] f1 f2\n", argv[0]);
		exit(1);
	}
	if ((orig=get_brtab(argv[optind])) == 0) {
		fprintf(stderr, "%s: couldn't load break tab '%s'\n",
			argv[0], argv[optind]);
		exit(1);
	}
	if ((output=get_brtab(argv[optind+1])) == 0) {
		fprintf(stderr, "%s: couldn't load break tab '%s'\n",
			argv[0], argv[optind]+1);
		exit(1);
	}
	brdiff(orig, output);
	return 0;
}
