/*-
 * mkbr - generate a breakpoint set for all reachable addresses.
 */

#include <libc.h>
#include "analyze.h"
#include "dbi.h"


int verbose = 0;
static uchar_t   *textimage;



find_text(gbl_info *sy, BreakTab *b)
{
	if (sy->kind & LMF_CODE && sy->name[strlen(sy->name)-1] == '_') {
		braflow(textimage, b, sy->addr.offset, 0, b->bh.bsize);
	}
}

setbrk(gbl_info *sy, BreakTab *b)
{
	if (sy->kind * LMF_CODE) {
		br_mark(b, sy->addr.offset);
	}
}

mkbtab(char *src, char *btab, void (*f)())
{
	FILE      *fp;
	lmfo      *obj;
	BreakTab  *b;
	uchar_t   *text;
	int        fd;
	ulong_t    moffs;
	symiter    it;
	gbl_info  *sym;

	if ((fp=fopen(src, "r")) == 0) {
		fprintf(stderr, "%s: cannot open\n", src);
		return -1;
	}

	if ((obj=lmf_open(fp)) == 0) {
		fclose(fp);
		return -1;
	}
	if ((b=new_brtab(src, obj->code_offset, obj->code_size)) == 0) {
		return -1;
	}
	text = 0;
	if (lmf_loadtext(obj, &text) == -1) {
		delete_brtab(b);
		lmf_close(obj);
		return -1;
	}
	textimage = text;
	if (lmf_symiter(obj, &it) == -1) {
		delete_brtab(b);
		lmf_close(obj);
		return -1;
	}	
	while (sym=lmf_symnext(&it)) {
		(*f)(sym, b);
	}
	if (put_brtab(btab, b) == -1) {
		delete_brtab(b);
		free(text);
		return -1;
	}
	delete_brtab(b);
	free(text);
	lmf_close(obj);
	return 0;
}

main(int argc, char **argv)
{
	int	c, i;
	char   *dir = ".";
	int    (*f)() = find_text;

	while ((c=getopt(argc, argv, "fvd:")) != -1) {
		switch (c) {
		case 'f': f = setbrk; break;
		case 'v': verbose = 1; break;
		case 'd': dir = optarg; break;
		}
	}
	for (i=optind; i < argc; i++) {
		char       pname[PATH_MAX+1];
		sprintf(pname, "%s/%s.brt", dir, basename(argv[i]));
		if (verbose) {
			printf("making table for '%s'\n", argv[i]);
		}
		if (mkbtab(argv[i], pname, f) == -1) {
			fprintf(stderr, "%s: couldn't analyze %s\n",
				argv[0], argv[i]);
		}
	}
	return 0;
}
