#ifndef dbi_h
#define dbi_h

#ifndef __STDIO_H_INCLUDED
#include <stdio.h>
#endif

#ifndef __TYPES_H_INCLUDED
#include <sys/types.h>
#endif

#ifndef __LMF_H_INCLUDED
#include <sys/lmf.h>
#endif



#pragma pack(1);

/*-
 *
 * These structures are shown in the accompany document
 * "WATCOM Debugging Information Format Version 4.0"
 * Copyright WATCOM International; April 30 1994
 *
 */
typedef unsigned short unsigned_16;
typedef unsigned long unsigned_32;
typedef unsigned char unsigned_8;

typedef struct master_dbg_header m_dbh;
typedef struct section_dbg_header sect_dbh;
typedef struct line_segment line_segment;
typedef struct line_info line_info;
typedef struct cue_entry cue_entry;
typedef struct mod_info mod_info;
typedef struct demand_info demand_info;

typedef struct gbl_info gbl_info;
typedef struct seg_info seg_info;
typedef struct addr_info addr_info;

typedef struct addr48_ptr addr48_ptr;


struct master_dbg_header {
	unsigned_16     signature;
	unsigned_8      exe_major_ver, exe_minor_ver;
	unsigned_8      obj_major_ver, obj_minor_ver;
	unsigned_16     lang_size;
	unsigned_16     segment_size;
	unsigned_32     debug_size;
};

struct section_dbg_header {
	unsigned_32     mod_offset;
	unsigned_32     gbl_offset;
	unsigned_32     addr_offset;
	unsigned_32     section_size;
	unsigned_16     section_id;
};

struct line_info {
	unsigned_16     line_number;
	unsigned_32     code_offset;
};

struct line_segment {
	unsigned_32     segment;
	unsigned_16     num;
	line_info       line[1];
};

struct cue_entry {
	unsigned_16     cue;
	unsigned_16     fno;
	unsigned_16     line;
	unsigned_16     column;
};

struct demand_info {
	unsigned_32     offset;
	unsigned_16     num_entries;
};

struct mod_info {
	unsigned_16     language;
	demand_info     locals;
	demand_info     types;
	demand_info     lines;
	unsigned_8      name[1];
};

struct  addr48_ptr {
	unsigned_32     offset;
	unsigned_16     segment;
};

struct gbl_info {
	addr48_ptr      addr;
	unsigned_16     mod_index;
	unsigned_8      kind;
#define LMF_STATIC       001
#define LMF_DATA         002
#define LMF_CODE         004
#define LMF_TYPEMASK     007
	unsigned_8      name[1];
};

struct addr_info {
	unsigned_32     size;
	unsigned_16     mod_index;
};

struct seg_info {
	addr48_ptr      addr;
	unsigned_16     num:15, nonsect:1;
	addr_info       sects[1];
};

#pragma pack(0);

typedef struct lmfo lmfo;

struct lmfo {
	unsigned    magic;

#define LMFO_MAGIC   0x5115
	FILE       *fp;
	ulong_t     code_offset;
	ulong_t     data_offset;
	ulong_t     code_size;
	ulong_t     data_size;
	ulong_t     bss_size;
	ulong_t     stack_size;
/*-
 * independently allocated:
 */
 	char      **langtab;
 	unsigned    nlang;
	mod_info  **modtab;
	unsigned    nmods;
	gbl_info  **globtab;
	unsigned    nglobs;
	seg_info  **segtab;
	unsigned    nsegs;
	char       *dbase;
	sect_dbh   *sect; /* is withing 'dbi' */
};

#define CHKMAGIC(lo)	\
	if ((lo)->magic != LMFO_MAGIC) { \
		errno = EPROGMISMATCH; \
		return -1; \
	}

extern  lmfo    *lmf_open(FILE *fp);
extern  int      lmf_close(lmfo *lo);
extern  int      lmf_fixname(unsigned char *name);
extern  void *   lmf_demand(lmfo *lo, demand_info *p, unsigned idx, unsigned *size);


extern  int lmf_a2l(lmfo *lo, ulong_t addr, char *fname, unsigned *lno);
extern  int lmf_a2m(lmfo *lo, ulong_t addr, char *fname, unsigned *offs);

extern  int lmf_print_modules(lmfo *lo, FILE *fp);
extern  int lmf_print_globs(lmfo *lo, FILE *fp);
extern  int lmf_print_segments(lmfo *lo, FILE *fp);

typedef struct {
	lmfo       *lmfo;
	int         index;
} symiter;

extern gbl_info *lmf_lookupdata(lmfo *lo, char *name);
extern gbl_info *lmf_lookuptext(lmfo *lo, char *name);
extern int       lmf_symiter(lmfo *lo, symiter *it);
extern gbl_info *lmf_symnext(symiter *it);


extern int       lmf_loadtext(lmfo *lo, char **text);
#endif
