/*-
 * brtab: set of functions to alloc, get, put, ... break point tables.
 */

#include <libc.h>
#include "analyze.h"


BreakTab *
new_brtab(char *src, ulong_t base, ulong_t size)
{
	BreakTab       *b;
	char           *x;

	if ((x=calloc(sizeof *x, sizeof *b+size)) == 0) {
		return 0;
	}
	b = (BreakTab *)x;
	b->bc = (uchar_t *)(b+1);
	b->bh.bsize = size;
	b->bh.base  = base;
	strcpy(b->bh.image, src);
	b->bh.magic = BMAGIC;
	return b;
}

int
delete_brtab(BreakTab *b)
{
	if (b) {
		b->bh.magic = -1;
		free(b);
	}
	return 0;
}

BreakTab  *
get_brtab(char *f)
{
	int	fd;
	BreakHeader       bh;
	BreakTab         *b;
	unsigned  i;
	if ((fd=open(f,O_RDONLY)) == -1) {
		fprintf(stderr, "can't open %s (%s)\n", f, strerror(errno));
		return 0;
	}
	if (read(fd, &bh, sizeof bh) != sizeof bh || bh.magic != BMAGIC) {
	    	fprintf(stderr, "invalid break table '%s'\n", f);
	    	close(fd);
	    	return 0;
	}
	if ((b=new_brtab(bh.image, bh.base, bh.bsize)) == 0) {
	    	fprintf(stderr, "couldn't alloc break table for '%s'\n", f);
		close(fd);
		return 0;
	}
	b->bh = bh;
	if (read(fd, b->bc, b->bh.bsize) != b->bh.bsize) {
		fprintf(stderr, "read failed '%s'\n", f);
		delete_brtab(b);
		return 0;
	}
	for (i=0; i < b->bh.bsize; i++) {
		b->bc[i] &= BrMask;
	}
	close(fd);
	return b;
}

int
put_brtab(char *f, BreakTab *brt)
{
	int	fd;
	unsigned i;
	if ((fd=creat(f,0666)) == -1) {
		fprintf(stderr, "can't creat %s (%s)\n", f, strerror(errno));
		return -1;
	}
	for (i=0; i < brt->bh.bsize; i++) {
		brt->bc[i] &= BrMask;
	}
	if (write(fd, &brt->bh, sizeof brt->bh) != sizeof brt->bh ||
	    write(fd, brt->bc, brt->bh.bsize) != brt->bh.bsize) {
	    	fprintf(stderr, "error writing '%s' (%s)\n", f, strerror(errno));
	    	return 1;
	}
	close(fd);
	return 0;
}
