/*-
 * ldf_load: load a program segment into memory.
 */

#include <libc.h>
#include <sys/lmf.h>
#include "ldfile.h"

static
find_seg(FILE *fp, int segno)
{
	struct _lmf_header    h;
	struct _lmf_data      d;

	while (fread(&h, sizeof h, 1, fp) == 1) {
		if (h.rec_type == _LMF_DATA_REC) {
			if (fread(&d, sizeof d, 1, fp) != 1)
				return -1;
			if (d.segment_index == segno) {
				fseek(fp, -(sizeof h+ sizeof d), SEEK_CUR);
				return 0;
			} else {
				fseek(fp, h.data_nbytes-sizeof d, SEEK_CUR);
			}
		} else {
			fseek(fp, h.data_nbytes, SEEK_CUR);
		}
	}
	return -1;
}

int
ldf_load(char *lfile, uchar_t **text, int segno)
{
	struct {
		struct _lmf_header     lh;
		struct _lmf_definition lr;
	} h;
	struct {
		struct _lmf_header     lh;
		struct _lmf_data       ld;
	} d;

	ulong_t       ssiz;
	FILE         *fp;
	if (ldf_ismagic(lfile) != 0 || (fp=fopen(lfile, "r")) == 0)
		return -1;
	if (fread(&h, sizeof h, 1, fp) != 1) {
		fclose(fp);
		return -1;
	}
	if (h.lh.data_nbytes < sizeof h.lr+segno*sizeof ssiz) {
		fclose(fp);
		return -1;
	}
	fseek(fp,  segno*sizeof ssiz, SEEK_CUR);
	if (fread(&ssiz, sizeof ssiz, 1, fp) != 1) {
		fclose(fp);
		return -1;
	}
	fseek(fp, sizeof h.lh + h.lh.data_nbytes, SEEK_SET);
/*-
 * next record: skip ahead until a matching segment data record is found.
 */
	if (find_seg(fp, segno) == -1) {
		fclose(fp);
		return -1;
	}
	if ((*text=calloc(sizeof **text, ssiz&0xfffffff)) == 0) {
		fclose(fp);
		return -1;
	}

/*-
 * read subsequent data records into the given offsets..
 */
	while (fread(&d, sizeof d, 1, fp) == 1 && d.lh.rec_type == _LMF_DATA_REC) {
		if (fread(*text+d.ld.offset, 1, d.lh.data_nbytes, fp) != d.lh.data_nbytes) {
			free(*text);
			fclose(fp);
			return -1;
		}
	}
	fclose(fp);
	return 0;
}


int
ldf_loadtext(char *lfile, uchar_t **text)
{

	return ldf_load(lfile, text, 0);
}
