
/*-
 * ldf_load: load a program segment into memory.
 */

#include <libc.h>
#include <sys/lmf.h>
#include "dbi.h"

static
seekseg(FILE *fp, int segno)
{
	struct _lmf_header    h;
	struct _lmf_data      d;

	while (fread(&h, sizeof h, 1, fp) == 1) {
		if (h.rec_type == _LMF_DATA_REC) {
			if (fread(&d, sizeof d, 1, fp) != 1)
				return -1;
			if (d.segment_index == segno) {
				fseek(fp, -(sizeof h+ sizeof d), SEEK_CUR);
				return 0;
			} else {
				fseek(fp, h.data_nbytes-sizeof d, SEEK_CUR);
			}
		} else {
			fseek(fp, h.data_nbytes, SEEK_CUR);
		}
	}
	return -1;
}

int
lmf_load(lmfo *lo, uchar_t **text, int segno)
{
	struct {
		struct _lmf_header     lh;
		struct _lmf_definition lr;
	} h;
	struct {
		struct _lmf_header     lh;
		struct _lmf_data       ld;
	} d;

	ulong_t       ssiz;
	ulong_t       nread = 0;

	fseek(lo->fp, 0, SEEK_SET);

	if (fread(&h, sizeof h, 1, lo->fp) != 1) {
		return -1;
	}
	if (h.lh.data_nbytes < sizeof h.lr+segno*sizeof ssiz) {
		return -1;
	}
	fseek(lo->fp,  segno*sizeof ssiz, SEEK_CUR);
	if (fread(&ssiz, sizeof ssiz, 1, lo->fp) != 1) {
		return -1;
	}
	ssiz &= 0xfffffff;
	fseek(lo->fp, sizeof h.lh + h.lh.data_nbytes, SEEK_SET);
/*-
 * next record: skip ahead until a matching segment data record is found.
 */
	if (seekseg(lo->fp, segno) == -1) {
		return -1;
	}
	if ((*text=calloc(sizeof **text, ssiz&0xfffffff)) == 0) {
		return -1;
	}

/*-
 * read subsequent data records into the given offsets..
 */
 	nread = 0;
	while (fread(&d, sizeof d, 1, lo->fp) == 1
	       && d.lh.rec_type == _LMF_DATA_REC
	       && d.ld.segment_index == segno) {
	       ulong_t    rsize = d.lh.data_nbytes - sizeof d.ld;
		if (fread(*text+d.ld.offset, 1, rsize, lo->fp) != rsize) {
			free(*text);
			return -1;
		}
		nread += rsize;
	}
	if (nread != ssiz) {
		fprintf(stderr, "warning: ssize was 0x%lx; read 0x%lx\n",
			ssiz, nread);
	}
	return 0;
}


int
lmf_loadtext(lmfo *lo, uchar_t **text)
{
	return lmf_load(lo, text, 0);
}
