
#include "dbi.h"
#include <errno.h>
#include <malloc.h>
#include <string.h>

/*-
 * technically, there can be many segment entries for each a real segment
 * as it is limitied to ~10000 symbols per segment.  The X8514 program
 * (biggest thing I could find) had about 1/10th that, so it's not
 * too worrying.  If it is a bother; I believe you just have to iterate
 * the segtable; however how you map that to code/data is a bit of a
 * bother.
 */

static seg_info *
lmf_getseg(lmfo *lo, unsigned isdata, ulong_t *addr)
{
	if (lo->nsegs > 2) {
		fprintf(stderr, "only expected 2 segments\n");
		return 0;
	}
	if (!isdata) {
		*addr  -= lo->code_offset;
		return lo->segtab[0];
	} else {
		*addr -= lo->data_offset;
		return lo->segtab[1];
	}
}


lmf_a2mod(lmfo *lo, ulong_t addr, mod_info **modi, unsigned *offs)
{
	unsigned        n;
	ulong_t         afix= addr;
	seg_info       *seg;

/*-
 * Given an address, find the defining module.
 */
	if ((seg=lmf_getseg(lo, 0, &afix)) == 0) {
		fprintf(stderr, "segment not found\n");
		return -1;
	}
	afix -= seg->addr.offset;
	for (n=0; n < seg->num && afix > seg->sects[n].size; n++)
		afix -= seg->sects[n].size;

	if (n == seg->num) {
#if 0
		fprintf(stderr, "exhausted segment\n");
#endif
		return -1;
	}
	*modi = lo->modtab[seg->sects[n].mod_index];
	*offs = afix;
	return 0;
}

int
lmf_a2m(lmfo *lo, ulong_t addr, char *fname, unsigned *offs)
{
	unsigned   mo;
	mod_info  *mop;
	if (lmf_a2mod(lo, addr, &mop, &mo) == 0) {
		strcpy(fname, mop->name);
		*offs = mo;
		return 0;
	}
	return -1;
}

int
lmf_a2l(lmfo *lo, ulong_t addr, char *fname, unsigned *lno)
{
	ulong_t         afix;
	mod_info       *mod;
	unsigned        i;

	if (lmf_a2mod(lo, addr, &mod, &afix) == -1) {
		return -1;
	}
/*-
 * From the module; find the line_table entry
 */ 
	strcpy(fname, mod->name);
	if (mod->lines.num_entries == 0) {
		*lno = -1;
#if 0
		fprintf(stderr, "no linenumber; got module\n");
#endif
		return 0;
	}
#ifdef DEBUGGING
fprintf(stderr, "looking for offset %x in %s\n",
	afix, mod->name);
#endif
	for (i=0; i < mod->lines.num_entries; i++) {
		line_segment       *lstrt, *ls;
		unsigned            j, size;
		if ((lstrt = lmf_demand(lo, &mod->lines, i, &size)) == 0)
			break;
		for (ls=lstrt; (char *)ls-(char *)lstrt < size;
			ls = (line_segment *)&ls->line[ls->num]) {
			for (j=0; j < ls->num; j++) {
				if (ls->line[j].code_offset > afix) {
					if (j == 0)
						*lno =ls->line[j].line_number-1;
					else
						*lno =ls->line[j].line_number;
					return 0;
				}
			}
		}
	}
#ifdef DEBUGGING
	fprintf(stderr, "not in linenumber tables!?!\n");
#endif
	return -1;
}
