#include "dbi.h"
#include <errno.h>
#include <malloc.h>
#include <unistd.h>

lmf_print_lines(lmfo *lo, mod_info *mo, FILE *fp)
{
	unsigned            i;
	line_segment       *lstrt, *ls;
	unsigned            j;

	fprintf(fp,"\tlines %u @0x%x\n",
			mo->lines.num_entries,
			mo->lines.offset);
	for (i=0; i < mo->lines.num_entries; i++) {
		unsigned       size;
		if ((lstrt=lmf_demand(lo, &mo->lines, i, &size)) == 0)
			break;
		for (ls=lstrt; (char *)ls-(char *)lstrt < size;
			ls = (line_segment *)&ls->line[ls->num]) {
#ifdef DEBUGGING
			fprintf(fp, "\t(segment=0x%x, #lines %u)\n",
				ls->segment, ls->num);
#endif
			for (j=0; j < ls->num; j++) {
				fprintf(fp, "\t\t%u @ 0x%x\n",
					ls->line[j].line_number,
					ls->line[j].code_offset);
			}
		}
	}
	return 0;
}

lmf_print_modules(lmfo *lo, FILE *fp)
{
	unsigned           n;

	CHKMAGIC(lo);
	for (n=0; n < lo->nmods; n++) {
		mod_info          *mo = lo->modtab[n];
		fprintf(fp,"%s:\n", mo->name);
		if (mo->locals.num_entries) {
			fprintf(fp,"\tlocals %u @0x%x\n",
				mo->locals.num_entries,
				mo->locals.offset);
		}
		if (mo->types.num_entries) {
			fprintf(fp,"\ttypes %u @0x%x\n",
				mo->types.num_entries,
				mo->types.offset);
		}
		if (mo->lines.num_entries) {
			lmf_print_lines(lo, mo, fp);
		}
	}
	return 0;
}

lmf_print_globs(lmfo *lo, FILE *fp)
{
	unsigned           n;

	CHKMAGIC(lo);
	for (n=0; n < lo->nmods; n++) {
		gbl_info          *go = lo->globtab[n];
		ulong_t            base = 0;
		char    ch;
		switch (go->kind & LMF_TYPEMASK) {
		case LMF_DATA:
			ch = 'D';
			base = lo->data_offset;
			break;
		case LMF_DATA|LMF_STATIC:
			ch = 'd';
			base = lo->data_offset;
			break;
		case LMF_CODE:
			ch = 'T';
			base = lo->code_offset;
			break;
		case LMF_CODE|LMF_STATIC:
			ch = 't';
			base = lo->code_offset;
			break;
		default:
			ch = '0'+(go->kind&7);
		}
		fprintf(fp, "%08.8lx %c %-32.32s%-20.20s\n",
			go->addr.offset + base,
			ch,
			go->name,
			basename(lo->modtab[go->mod_index]->name));
	}
	return 0;
}

int
lmf_print_segments(lmfo *lo, FILE *fp)
{
	unsigned        n;
	CHKMAGIC(lo);
	for (n=0; n < lo->nsegs; n++) {
		seg_info	*s0 = lo->segtab[n];
		unsigned         i;
		fprintf(fp,"%d: addr 0x%x:0x%x %u sects %s\n",
			n,
			s0->addr.segment,
			s0->addr.offset,
			s0->num,
			s0->nonsect? "NonSect" :"");
			for (i=0; i < s0->num; i++) {
				fprintf(fp, "\t0x%lx %s\n",
					s0->sects[i].size,
					lo->modtab[s0->sects[i].mod_index]->name);
			}
	}
	return 0;
}
