#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "analyze.h"


#ifndef _argused
#define _argused(a) ((a)=(a))
#endif

typedef signed char schar_t;

int     disasm_mode = 1;

static int      mod;
static int      rm;
static int      reg;
static int      dflag;		/* dflag = sizeof data */
static int      aflag;		/* aflag = sizeof addr */
static int      prefixes;       /* current prefix set */

extern int         INIT(char *, unsigned);
extern uchar_t     CURRENT(void);
extern uchar_t     PREVIOUS(void);
extern uchar_t     ADVANCE(void);
extern unsigned    OFFSET(void);
extern unsigned    BEGIN(void);


static ulong_t
get32()
{
	ulong_t          x = 0;

	x = ADVANCE() & 0xff;
	x |= (ADVANCE() & 0xff) << 8;
	x |= (ADVANCE() & 0xff) << 16;
	x |= (ADVANCE() & 0xff) << 24;
	return x;
}

static ulong_t
get16()
{
	ulong_t          x = 0;

	x = ADVANCE() & 0xff;
	x |= (ADVANCE() & 0xff) << 8;
	return x;
}

#define PREFIX_REPZ  (1 << 0)
#define PREFIX_REPNZ (1 << 1)
#define PREFIX_LOCK  (1 << 2)
#define PREFIX_CS    (1 << 3)
#define PREFIX_SS    (1 << 4)
#define PREFIX_DS    (1 << 5)
#define PREFIX_ES    (1 << 6)
#define PREFIX_FS    (1 << 7)
#define PREFIX_GS    (1 << 8)
#define PREFIX_DATA  (1 << 9)
#define PREFIX_ADR   (1 << 10)
#define PREFIX_FWAIT (1 << 11)

/* capital letters in template are macros */
static
putop(template, obufp)
	char           *template;
	char           *obufp;
{
	char           *p;

	for (p = template; *p; p++) {
		switch (*p) {
		default:
			*obufp++ = *p;
			break;
		case 'C':	/* For jcxz/jecxz */
			if (aflag == 0)
				*obufp++ = 'e';
			break;
		case 'N':
			if ((prefixes & PREFIX_FWAIT) == 0)
				*obufp++ = 'n';
			break;
		case 'S':
			/* operand size flag */
			if (dflag)
				*obufp++ = 'l';
			else
				*obufp++ = 'w';
			break;
		}
	}
	*obufp = 0;
}

char *
puthex(char *buf, long val)
{
	*buf++ = '0';
	*buf++ = 'x';
	return buf + strlen(ltoa(val, buf, 16));
}



#define Eb      'e', b_mode
#define indEb 'E', b_mode
#define Gb      'g', b_mode
#define Ev      'e', v_mode
#define indEv 'E', v_mode
#define Ew      'e', w_mode
#define Ma      'e', v_mode
#define M       'e', 0
#define Mp      'e', 0		/* ? */
#define Gv      'g', v_mode
#define Gw      'g', w_mode
#define Rw      'r', w_mode
#define Rd      'r', d_mode
#define Ib      'i', b_mode
#define sIb     'I', b_mode	/* sign extened byte */
#define Iv      'i', v_mode
#define Iw      'i', w_mode
#define Jb      'j', b_mode
#define Jv      'j', v_mode
#define ONE     '1', 0
#define Cd      'c', d_mode
#define Dd      'd', d_mode
#define Td      't', d_mode
#define eAX     'R', eAX_reg
#define eBX     'R', eBX_reg
#define eCX     'R', eCX_reg
#define eDX     'R', eDX_reg
#define eSP     'R', eSP_reg
#define eBP     'R', eBP_reg
#define eSI     'R', eSI_reg
#define eDI     'R', eDI_reg
#define AL      'R', al_reg
#define CL      'R', cl_reg
#define DL      'R', dl_reg
#define BL      'R', bl_reg
#define AH      'R', ah_reg
#define CH      'R', ch_reg
#define DH      'R', dh_reg
#define BH      'R', bh_reg
#define AX      'R', ax_reg
#define DX      'R', dx_reg
#define indDX   'R', ind_dx_reg
#define Sw      'S', w_mode
#define Ap      'D', lptr
#define Av      'D', v_mode
#define Ob      'O', b_mode
#define Ov      'O', v_mode
#define Xb      'm', b_mode
#define Xv      'm', v_mode
#define Yb      'M', b_mode
#define Yv      'M', v_mode

#define es      'R', es_reg
#define ss      'R', ss_reg
#define cs      'R', cs_reg
#define ds      'R', ds_reg
#define fs      'R', fs_reg
#define gs      'R', gs_reg

#define b_mode 1
#define v_mode 2
#define w_mode 3
#define d_mode 4

#define es_reg 100
#define cs_reg 101
#define ss_reg 102
#define ds_reg 103
#define fs_reg 104
#define gs_reg 105
#define eAX_reg 107
#define eCX_reg 108
#define eDX_reg 109
#define eBX_reg 110
#define eSP_reg 111
#define eBP_reg 112
#define eSI_reg 113
#define eDI_reg 114

#define lptr 115

#define al_reg 116
#define cl_reg 117
#define dl_reg 118
#define bl_reg 119
#define ah_reg 120
#define ch_reg 121
#define dh_reg 122
#define bh_reg 123

#define ax_reg 124
#define cx_reg 125
#define dx_reg 126
#define bx_reg 127
#define sp_reg 128
#define bp_reg 129
#define si_reg 130
#define di_reg 131

#define ind_dx_reg 150

#define GRP1b     NULL, NULL, 0
#define GRP1S     NULL, NULL, 1
#define GRP1Ss    NULL, NULL, 2
#define GRP2b     NULL, NULL, 3
#define GRP2S     NULL, NULL, 4
#define GRP2b_one NULL, NULL, 5
#define GRP2S_one NULL, NULL, 6
#define GRP2b_cl  NULL, NULL, 7
#define GRP2S_cl  NULL, NULL, 8
#define GRP3b     NULL, NULL, 9
#define GRP3S     NULL, NULL, 10
#define GRP4      NULL, NULL, 11
#define GRP5      NULL, NULL, 12
#define GRP6      NULL, NULL, 13
#define GRP7      NULL, NULL, 14
#define GRP8      NULL, NULL, 15

#define FLOATCODE 50
#define FLOAT     NULL, NULL, FLOATCODE

struct dis386 {
	char           *name;
	char		op1, bytemode1;
	char		op2, bytemode2;
	char		op3, bytemode3;
};


#define BAD   "(bad)"

#define AAA   "aaa"
#define AAD   "aad"
#define ADC   "adc"
#define ADD   "add"
#define AAM   "aam"
#define AAS   "aas"
#define AND   "and"
#define ARPL  "arpl"

#define BOUND "bound"
#define BSF   "bsf"
#define BSR   "bsr"
#define BSWAP "bswap"       /* not in tables */
#define BT    "bt"
#define BTC   "btc"
#define BTR   "btr"
#define BTS   "bts"

#define CALL  "call"
#define CALLF "lcall"
#define CBW   "cbw"
#define CLC   "clc"
#define CLD   "cld"
#define CLI   "cli"
#define CLTS  "clts"
#define CMC   "cmc"

#define PUSH  "push"
#define POP   "pop"




static struct dis386   dis386[] = {
 /* 00 */
 {ADD, Eb, Gb}, {ADD, Ev, Gv},  {ADD, Gb, Eb}, {ADD, Gv, Ev},
 {ADD, AL, Ib}, {ADD, eAX, Iv}, {PUSH, es},    {POP, es},
 /* 08 */
 {"orb", Eb, Gb},  {"orS", Ev, Gv},   {"orb", Gb, Eb},  {"orS", Gv, Ev},
 {"orb", AL, Ib},  {"orS", eAX, Iv} , {PUSH, cs},    {BAD},
/* 0x0f extended opcode escape */
 /* 10 */
 {ADC, Eb, Gb}, {ADC, Ev, Gv},  {ADC, Gb, Eb}, {ADC, Gv, Ev},
 {ADC, AL, Ib}, {ADC, eAX, Iv}, {PUSH, ss},    {POP, ss},
 /* 18 */
 {"sbbb", Eb, Gb}, {"sbbS", Ev, Gv},  {"sbbb", Gb, Eb}, {"sbbS", Gv, Ev},
 {"sbbb", AL, Ib}, {"sbbS", eAX, Iv}, {PUSH, ds},    {POP, ds},
 /* 20 */
 {AND, Eb, Gb}, {AND, Ev, Gv},  {AND, Gb, Eb}, {AND, Gv, Ev},
 {AND, AL, Ib}, {AND, eAX, Iv}, {"es:"},          {"daa"},
 /* 28 */
 {"subb", Eb, Gb}, {"subS", Ev, Gv},  {"subb", Gb, Eb}, {"subS", Gv, Ev},
 {"subb", AL, Ib}, {"subS", eAX, Iv}, {"cs:"},          {"das"},
 /* 30 */
 {"xorb", Eb, Gb}, {"xorS", Ev, Gv},  {"xorb", Gb, Eb}, {"xorS", Gv, Ev},
 {"xorb", AL, Ib}, {"xorS", eAX, Iv}, {"ss:"},          {AAA},
 /* 38 */
 {"cmpb", Eb, Gb}, {"cmpS", Ev, Gv},  {"cmpb", Gb, Eb}, {"cmpS", Gv, Ev},
 {"cmpb", AL, Ib}, {"cmpS", eAX, Iv}, {"ds:"},          {AAS},
 /* 40 */
 {"incS", eAX},    {"incS", eCX},     {"incS", eDX},    {"incS", eBX},
 {"incS", eSP},    {"incS", eBP},     {"incS", eSI},    {"incS", eDI},
 /* 48 */
 {"decS", eAX},    {"decS", eCX},     {"decS", eDX},    {"decS", eBX},
 {"decS", eSP},    {"decS", eBP},     {"decS", eSI},    {"decS", eDI},
 /* 50 */
 {PUSH, eAX},   {PUSH, eCX},    {PUSH, eDX},   {PUSH, eBX},
 {PUSH, eSP},   {PUSH, eBP},    {PUSH, eSI},   {PUSH, eDI},
 /* 58 */
 {POP, eAX},    {POP, eCX},     {POP, eDX},    {POP, eBX},
 {POP, eSP},    {POP, eBP},     {POP, eSI},    {POP, eDI},
 /* 60 */
 {"pushad"},        {"popad"},          {BOUND, Gv, Ma},{ARPL, Ew, Gw},
 {"fs:"},          {"gs:"},           {"opr16:"},       {"addr16"},
 /* 68 -- NOTE #68 is wrong in 386 boook */
 {PUSH, Iv},    {"imulS", Gv, Ev, Iv},{PUSH, sIb},{"imulS", Gv, Ev, Ib},
 {"insb", Yb, indDX},{"insS", Yv, indDX},{"outsb", indDX, Xb},{"outsS", indDX, Xv},
 /* 70 */
 {"jo", Jb},       {"jno", Jb},       {"jb", Jb},       {"jae", Jb},
 {"je", Jb},       {"jne", Jb},       {"jbe", Jb},      {"ja", Jb},
 /* 78 */
 {"js", Jb},       {"jns", Jb},       {"jp", Jb},       {"jnp", Jb},
 {"jl", Jb},       {"jnl", Jb},       {"jle", Jb},      {"jg", Jb},
 /* 80 */
 {GRP1b},          {GRP1S},           {BAD},        {GRP1Ss},
 {"testb", Eb, Gb},{"testS", Ev, Gv}, {"xchgb", Eb, Gb},{"xchgS", Ev, Gv},
 /* 88 */
 {"movb", Eb, Gb}, {"movS", Ev, Gv},  {"movb", Gb, Eb}, {"movS", Gv, Ev},
 {"movw", Ew, Sw}, {"leaS", Gv, M},   {"movw", Sw, Ew}, {POP, Ev},
 /* 90 */
 {"nop"},          {"xchgS", eCX, eAX},{"xchgS", eDX, eAX},{"xchgS", eBX, eAX},
 {"xchgS", eSP, eAX},{"xchgS", eBP, eAX},{"xchgS", eSI, eAX},{"xchgS", eDI, eAX},
 /* 98 */
 {CBW},         {"cltd"},          {CALLF, Ap},    {"fwait"},
 {PUSH},        {POP},          {"sahf"},         {"lahf"},
 /* a0 */
 {"movb", AL, Ob}, {"movS", eAX, Ov}, {"movb", Ob, AL}, {"movS", Ov, eAX},
 {"movsb", Yb, Xb},{"movsS", Yv, Xv}, {"cmpsb", Yb, Xb},{"cmpsS", Yv, Xv},
 /* a8 */
 {"testb", AL, Ib},{"testS", eAX, Iv},{"stosb", Yb, AL},{"stosS", Yv, eAX},
 {"lodsb", AL, Xb},{"lodsS", eAX, Xv},{"scasb", AL, Xb},{"scasS", eAX, Xv},
 /* b0 */
 {"movb", AL, Ib}, {"movb", CL, Ib},  {"movb", DL, Ib}, {"movb", BL, Ib},
 {"movb", AH, Ib}, {"movb", CH, Ib},  {"movb", DH, Ib}, {"movb", BH, Ib},
 /* b8 */
 {"movS", eAX, Iv},{"movS", eCX, Iv}, {"movS", eDX, Iv},{"movS", eBX, Iv},
 {"movS", eSP, Iv},{"movS", eBP, Iv}, {"movS", eSI, Iv},{"movS", eDI, Iv},
 /* c0 */
 {GRP2b},          {GRP2S},           {"ret", Iw},      {"ret"},
 {"lesS", Gv, Mp}, {"ldsS", Gv, Mp},  {"movb", Eb, Ib}, {"movS", Ev, Iv},
 /* c8 */
 {"enter", Iw, Ib},{"leave"},         {"lret", Iw},     {"lret"},
 {"int3"},         {"int", Ib},       {"into"},         {"iret"},
 /* d0 */
 {GRP2b_one},      {GRP2S_one},       {GRP2b_cl},       {GRP2S_cl},
 {AAM, Ib},      {AAD, Ib},       {BAD},        {"xlat"},
 /* d8 */
 {FLOAT},          {FLOAT},           {FLOAT},          {FLOAT},
 {FLOAT},          {FLOAT},           {FLOAT},          {FLOAT},
 /* e0 */
 {"loopne", Jb},   {"loope", Jb},     {"loop", Jb},     {"jCcxz", Jb},
 {"inb", AL, Ib},  {"inS", eAX, Ib},  {"outb", Ib, AL}, {"outS", Ib, eAX},
 /* e8 */
 {CALL, Av},     {"jmp", Jv},       {"ljmp", Ap},     {"jmp", Jb},
 {"inb", AL, indDX},{"inS", eAX, indDX},{"outb", indDX, AL},{"outS", indDX, eAX},
 /* f0 */
 {"lock:"},        {BAD},         {"repne:"},       {"repz:"},
 {"hlt"},          {CMC},           {GRP3b},          {GRP3S},
 /* f8 */
 {CLC},          {"stc"},           {CLI},          {"sti"},
 {CLD},          {"std"},           {GRP4},           {GRP5},
};


static struct dis386 dis386_twobyte[] = {
 {BAD},       {GRP6},             {GRP7},           {"larS", Gv, Ew},
 {"lslS", Gv, Ew},{CLTS},           {"movl", Rd, Cd}, {"movl", Rd, Dd},
 {"movl", Cd, Rd},{"movl", Dd, Rd},   {"movl", Rd, Td}, {"movl", Td, Rd},
 {"jo", Jv},      {"jno", Jv},        {"jb", Jv},       {"jae", Jv},
 {"je", Jv},      {"jne", Jv},        {"jbe", Jv},      {"ja", Jv},
 {"js", Jv},      {"jns", Jv},        {"jp", Jv},       {"jnp", Jv},
 {"jl", Jv},      {"jge", Jv},        {"jle", Jv},      {"jg", Jv},
 {"seto", Eb},    {"setno", Eb},      {"setb", Eb},     {"setae", Eb},
 {"sete", Eb},    {"setne", Eb},      {"setbe", Eb},    {"seta", Eb},
 {"sets", Eb},    {"setns", Eb},      {"setp", Eb},     {"setnp", Eb},
 {"setl", Eb},    {"setge", Eb},      {"setle", Eb},    {"setg", Eb},
 {PUSH, fs},   {POP, fs},       {BT, Ev, Gv},  {"shldS", Ev, Gv, Ib},
 {"shldS", Ev, Gv, CL},
 {PUSH, gs},   {POP, gs},       {BTS, Ev, Gv},
 {"shrdS", Ev, Gv, Ib},
 {"shrdS", Ev, Gv, CL},
 {"imulS", Gv, Ev},{"lssS", Gv, Mp},  {BTR, Ev, Gv}, {"lfsS", Gv, Mp},
 {"lgsS", Gv, Mp}, {"movzbS", Gv, Eb},{"movzwS", Gv, Ew},{GRP8},
 {BTC, Ev, Gv}, {BSF, Gv, Ev},  {BSR, Gv, Ev}, {"movsbS", Gv, Eb},
 {"movswS", Gv, Ew}
};

static uchar_t dis_twoind[] = {
	1,   2,   3,   4,   0,   0,   5,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        6,   7,   8,   9,  10,   0,  11,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,
/* 40 */
        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,
/* 80 */
       12,  13,  14,  15,  16,  17,  18,  19,
       20,  21,  22,  23,  24,  25,  26,  27,

       28,  29,  30,  31,  32,  33,  34,  35,
       36,  37,  38,  39,  40,  41,  42,  43,
 
       44,  45,   0,  46,  47,  48,   0,   0,
       49,  50,   0,  51,  52,  53,   0,  54,

        0,   0,  55,  56,  57,  58,  59,  60,
        0,   0,  61,  62,  63,  64,  65,  66,
/* c0 */
        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,

        0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,
};


#define rnum_eax 0
#define rnum_ecx 1
#define rnum_edx 2
#define rnum_ebx 3
#define rnum_esp 4
#define rnum_ebp 5
#define rnum_esi 6
#define rnum_edi 7

#define snum_es 0
#define snum_cs 1
#define snum_ss 2
#define snum_ds 3
#define snum_fs 4
#define snum_gs 5

static char    *names32[] = {
	"%eax", "%ecx", "%edx", "%ebx", "%esp", "%ebp", "%esi", "%edi",
};
static char    *names16[] = {
	"%ax", "%cx", "%dx", "%bx", "%sp", "%bp", "%si", "%di",
};
static char    *names8[] = {
	"%al", "%cl", "%dl", "%bl", "%ah", "%ch", "%dh", "%bh",
};
static char    *names_seg[] = {
	"%es", "%cs", "%ss", "%ds", "%fs", "%gs", "%?", "%?",
};






static char    *
segname(int segno)
{
	return names_seg[segno - es_reg] + disasm_mode;
}

static char    *
regname(int regno, int is32)
{
	return (is32 ? names32 : names16)[regno] + disasm_mode;
}

static struct dis386   grps[][8] = {
	/* GRP1b */
	{
		{ADD, Eb, Ib},
		{"orb", Eb, Ib},
		{ADC, Eb, Ib},
		{"sbbb", Eb, Ib},
		{AND, Eb, Ib},
		{"subb", Eb, Ib},
		{"xorb", Eb, Ib},
		{"cmpb", Eb, Ib}
	},
	/* GRP1S */
	{
		{ADD, Ev, Iv},
		{"orS", Ev, Iv},
		{ADC, Ev, Iv},
		{"sbbS", Ev, Iv},
		{AND, Ev, Iv},
		{"subS", Ev, Iv},
		{"xorS", Ev, Iv},
		{"cmpS", Ev, Iv}
	},
	/* GRP1Ss */
	{
		{ADD, Ev, sIb},
		{"orS", Ev, sIb},
		{ADC, Ev, sIb},
		{"sbbS", Ev, sIb},
		{AND, Ev, sIb},
		{"subS", Ev, sIb},
		{"xorS", Ev, sIb},
		{"cmpS", Ev, sIb}
	},
	/* GRP2b */
	{
		{"rolb", Eb, Ib},
		{"rorb", Eb, Ib},
		{"rclb", Eb, Ib},
		{"rcrb", Eb, Ib},
		{"shlb", Eb, Ib},
		{"shrb", Eb, Ib},
		{BAD},
		{"sarb", Eb, Ib},
	},
	/* GRP2S */
	{
		{"rolS", Ev, Ib},
		{"rorS", Ev, Ib},
		{"rclS", Ev, Ib},
		{"rcrS", Ev, Ib},
		{"shlS", Ev, Ib},
		{"shrS", Ev, Ib},
		{BAD},
		{"sarS", Ev, Ib},
	},
	/* GRP2b_one */
	{
		{"rolb", Eb},
		{"rorb", Eb},
		{"rclb", Eb},
		{"rcrb", Eb},
		{"shlb", Eb},
		{"shrb", Eb},
		{BAD},
		{"sarb", Eb},
	},
	/* GRP2S_one */
	{
		{"rolS", Ev},
		{"rorS", Ev},
		{"rclS", Ev},
		{"rcrS", Ev},
		{"shlS", Ev},
		{"shrS", Ev},
		{BAD},
		{"sarS", Ev},
	},
	/* GRP2b_cl */
	{
		{"rolb", Eb, CL},
		{"rorb", Eb, CL},
		{"rclb", Eb, CL},
		{"rcrb", Eb, CL},
		{"shlb", Eb, CL},
		{"shrb", Eb, CL},
		{BAD},
		{"sarb", Eb, CL},
	},
	/* GRP2S_cl */
	{
		{"rolS", Ev, CL},
		{"rorS", Ev, CL},
		{"rclS", Ev, CL},
		{"rcrS", Ev, CL},
		{"shlS", Ev, CL},
		{"shrS", Ev, CL},
		{BAD},
		{"sarS", Ev, CL}
	},
	/* GRP3b */
	{
		{"testb", Eb, Ib},
		{BAD, Eb},
		{"notb", Eb},
		{"negb", Eb},
		{"mulb", AL, Eb},
		{"imulb", AL, Eb},
		{"divb", AL, Eb},
		{"idivb", AL, Eb}
	},
	/* GRP3S */
	{
		{"testS", Ev, Iv},
		{BAD},
		{"notS", Ev},
		{"negS", Ev},
		{"mulS", eAX, Ev},
		{"imulS", eAX, Ev},
		{"divS", eAX, Ev},
		{"idivS", eAX, Ev},
	},
	/* GRP4 */
	{
		{"incb", Eb},
		{"decb", Eb},
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{BAD},
	},
	/* GRP5 */
	{
		{"incS", Ev},
		{"decS", Ev},
		{CALL, indEv},
		{CALLF, indEv},
		{"jmp", indEv},
		{"ljmp", indEv},
		{PUSH, Ev},
		{BAD},
	},
	/* GRP6 */
	{
		{"sldt", Ew},
		{"str", Ew},
		{"lldt", Ew},
		{"ltr", Ew},
		{"verr", Ew},
		{"verw", Ew},
		{BAD},
		{BAD}
	},
	/* GRP7 */
	{
		{"sgdt", Ew},
		{"sidt", Ew},
		{"lgdt", Ew},
		{"lidt", Ew},
		{"smsw", Ew},
		{BAD},
		{"lmsw", Ew},
		{BAD},
	},
	/* GRP8 */
	{
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{BT, Ev, Ib},
		{BTS, Ev, Ib},
		{BTR, Ev, Ib},
		{BTC, Ev, Ib},
	}
};

static char           *float_mem[] = {
	/* d8 */
	"fadds",
	"fmuls",
	"fcoms",
	"fcomps",
	"fsubs",
	"fsubrs",
	"fdivs",
	"fdivrs",
	/* d9 */
	"flds",
	BAD,
	"fsts",
	"fstps",
	"fldenv",
	"fldcw",
	"fNstenv",
	"fNstcw",
	/* da */
	"fiaddl",
	"fimull",
	"ficoml",
	"ficompl",
	"fisubl",
	"fisubrl",
	"fidivl",
	"fidivrl",
	/* db */
	"fildl",
	BAD,
	"fistl",
	"fistpl",
	BAD,
	"fldt",
	BAD,
	"fstpt",
	/* dc */
	"faddl",
	"fmull",
	"fcoml",
	"fcompl",
	"fsubl",
	"fsubrl",
	"fdivl",
	"fdivrl",
	/* dd */
	"fldl",
	BAD,
	"fstl",
	"fstpl",
	"frstor",
	BAD,
	"fNsave",
	"fNstsw",
	/* de */
	"fiadd",
	"fimul",
	"ficom",
	"ficomp",
	"fisub",
	"fisubr",
	"fidiv",
	"fidivr",
	/* df */
	"fild",
	BAD,
	"fist",
	"fistp",
	"fbld",
	"fildll",
	"fbstp",
	"fistpll",
};

#define ST 'f', 0
#define STi 'F', 0

#define FGRPd9_2 NULL, NULL, 0
#define FGRPd9_4 NULL, NULL, 1
#define FGRPd9_5 NULL, NULL, 2
#define FGRPd9_6 NULL, NULL, 3
#define FGRPd9_7 NULL, NULL, 4
#define FGRPda_5 NULL, NULL, 5
#define FGRPdb_4 NULL, NULL, 6
#define FGRPde_3 NULL, NULL, 7
#define FGRPdf_4 NULL, NULL, 8

static struct dis386   float_reg[][8] = {
	/* d8 */
	{
		{"fadd", ST, STi},
		{"fmul", ST, STi},
		{"fcom", STi},
		{"fcomp", STi},
		{"fsub", ST, STi},
		{"fsubr", ST, STi},
		{"fdiv", ST, STi},
		{"fdivr", ST, STi},
	},
	/* d9 */
	{
		{"fld", STi},
		{"fxch", STi},
		{FGRPd9_2},
		{BAD},
		{FGRPd9_4},
		{FGRPd9_5},
		{FGRPd9_6},
		{FGRPd9_7},
	},
	/* da */
	{
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{FGRPda_5},
		{BAD},
		{BAD},
	},
	/* db */
	{
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{FGRPdb_4},
		{BAD},
		{BAD},
		{BAD},
	},
	/* dc */
	{
		{"fadd", STi, ST},
		{"fmul", STi, ST},
		{BAD},
		{BAD},
		{"fsub", STi, ST},
		{"fsubr", STi, ST},
		{"fdiv", STi, ST},
		{"fdivr", STi, ST},
	},
	/* dd */
	{
		{"ffree", STi},
		{BAD},
		{"fst", STi},
		{"fstp", STi},
		{"fucom", STi},
		{"fucomp", STi},
		{BAD},
		{BAD},
	},
	/* de */
	{
		{"faddp", STi, ST},
		{"fmulp", STi, ST},
		{BAD},
		{FGRPde_3},
		{"fsubp", STi, ST},
		{"fsubrp", STi, ST},
		{"fdivp", STi, ST},
		{"fdivrp", STi, ST},
	},
	/* df */
	{
		{BAD},
		{BAD},
		{BAD},
		{BAD},
		{FGRPdf_4},
		{BAD},
		{BAD},
		{BAD},
	},
};


static char           *fgrps[][8] = {
	/* d9_2  0 */
	{
		"fnop", BAD, BAD, BAD, BAD, BAD, BAD, BAD,
	},

	/* d9_4  1 */
	{
		"fchs", "fabs", BAD, BAD, "ftst", "fxam", BAD, BAD,
	},

	/* d9_5  2 */
	{
		"fld1", "fldl2t", "fldl2e", "fldpi", "fldlg2", "fldln2", "fldz", BAD,
	},

	/* d9_6  3 */
	{
		"f2xm1", "fyl2x", "fptan", "fpatan", "fxtract", "fprem1", "fdecstp", "fincstp",
	},

	/* d9_7  4 */
	{
		"fprem", "fyl2xp1", "fsqrt", "fsincos", "frndint", "fscale", "fsin", "fcos",
	},

	/* da_5  5 */
	{
		BAD, "fucompp", BAD, BAD, BAD, BAD, BAD, BAD,
	},

	/* db_4  6 */
	{
		"feni(287 only)", "fdisi(287 only)", "fNclex", "fNinit",
		"fNsetpm(287 only)", BAD, BAD, BAD,
	},

	/* de_3  7 */
	{
		BAD, "fcompp", BAD, BAD, BAD, BAD, BAD, BAD,
	},

	/* df_4  8 */
	{
		"fNstsw", BAD, BAD, BAD, BAD, BAD, BAD, BAD,
	},
};


static
add_prefix(char *p)
{
	if (prefixes & PREFIX_CS)
		strcat(strcat(p, names_seg[snum_cs] + disasm_mode), ":");
	if (prefixes & PREFIX_DS)
		strcat(strcat(p, names_seg[snum_ds] + disasm_mode), ":");
	if (prefixes & PREFIX_SS)
		strcat(strcat(p, names_seg[snum_ss] + disasm_mode), ":");
	if (prefixes & PREFIX_ES)
		strcat(strcat(p, names_seg[snum_es] + disasm_mode), ":");
	if (prefixes & PREFIX_FS)
		strcat(strcat(p, names_seg[snum_fs] + disasm_mode), ":");
	if (prefixes & PREFIX_GS)
		strcat(strcat(p, names_seg[snum_gs] + disasm_mode), ":");
}

static
OP_E(int bytemode, char *obuff)
{
	int             disp;
	int             havesib;
	int             base;
	int             index;
	int             scale;
	int             havebase;

	/* skip mod/rm byte */
	ADVANCE();

	havesib = 0;
	havebase = 0;
	disp = 0;

	if (mod == 3) {
		switch (bytemode) {
		case b_mode:
			strcpy(obuff,names8[rm] + disasm_mode);
			break;
		case w_mode:
			strcpy(obuff, names16[rm] + disasm_mode);
			break;
		case v_mode:
			if (dflag)
				strcpy(obuff, names32[rm] + disasm_mode);
			else
				strcpy(obuff, names16[rm] + disasm_mode);
			break;
		default:
			strcpy(obuff, "<bad dis table>");
			break;
		}
		return 0;
	}
	add_prefix(obuff);
	if (rm == 4) {
		havesib = 1;
		havebase = 1;
		scale = (CURRENT() >> 6) & 3;
		index = (CURRENT() >> 3) & 7;
		base = ADVANCE() & 7;
	}
	switch (mod) {
	case 0:
		switch (rm) {
		case 4:
			/* implies havesib and havebase */
			if (base == 5) {
				havebase = 0;
				disp = get32();
			}
			break;
		case 5:
			disp = get32();
			break;
		default:
			havebase = 1;
			base = rm;
			break;
		}
		break;
	case 1:
		disp = (schar_t) ADVANCE();
		if (rm != 4) {
			havebase = 1;
			base = rm;
		}
		break;
	case 2:
		disp = get32();
		if (rm != 4) {
			havebase = 1;
			base = rm;
		}
		break;
	}

	if (mod != 0 || rm == 5 || (havesib && base == 5)) {
		ltoa(disp, obuff, 10);
	}
	if (havebase || havesib) {
		char   *p = obuff+strlen(obuff);
		*p++ = '(';
		if (havebase) {
			strcpy(p, names32[base] + disasm_mode);
			p += strlen(p);
		}
		if (havesib) {
			if (index != 4) {
				*p++ = ',';
				strcpy(p, names32[index] + disasm_mode);
				p += strlen(p);
			}
			*p++ = ',';
			ltoa(1<<scale, p, 10);
			p += strlen(p);
		}
		*p++ = ')';
		*p = '\0';
	}
	return 0;
}

static
OP_G(int bytemode, char *obuff)
{
	char	*s;
	switch (bytemode) {
	case b_mode:
		s = (names8[reg] + disasm_mode);
		break;
	case w_mode:
		s = (names16[reg] + disasm_mode);
		break;
	case d_mode:
		s = (names32[reg] + disasm_mode);
		break;
	case v_mode:
		if (dflag)
			s = (names32[reg] + disasm_mode);
		else
			s = (names16[reg] + disasm_mode);
		break;
	default:
		s = "<internal disassembler error>";
		break;
	}
	strcpy(obuff, s);
	return 0;
}

static
OP_REG(int code, char *obuff)
{
	char           *s;

	switch (code) {
	case ind_dx_reg:
		s = disasm_mode ? "(dx)" : "(%dx)";
		break;
	case ax_reg:
	case cx_reg:
	case dx_reg:
	case bx_reg:
	case sp_reg:
	case bp_reg:
	case si_reg:
	case di_reg:
		s = regname(code - ax_reg, 0);
		break;
	case es_reg:
	case ss_reg:
	case cs_reg:
	case ds_reg:
	case fs_reg:
	case gs_reg:
		s = segname(code);
		break;
	case al_reg:
	case ah_reg:
	case cl_reg:
	case ch_reg:
	case dl_reg:
	case dh_reg:
	case bl_reg:
	case bh_reg:
		s = names8[code - al_reg] + disasm_mode;
		break;
	case eAX_reg:
	case eCX_reg:
	case eDX_reg:
	case eBX_reg:
	case eSP_reg:
	case eBP_reg:
	case eSI_reg:
	case eDI_reg:
		if (dflag)
			s = names32[code - eAX_reg] + disasm_mode;
		else
			s = names16[code - eAX_reg] + disasm_mode;
		break;
	default:
		s = "<internal disassembler error>";
		break;
	}
	strcpy(obuff, s);
	return 0;
}

static
OP_I(int bytemode, char *obuff)
{
	int             op;

	switch (bytemode) {
	case b_mode:
		op = ADVANCE() & 0xff;
		break;
	case v_mode:
		op = dflag? get32() : get16();
		break;
	case w_mode:
		op = get16();
		break;
	default:
		strcpy(obuff, "<internal disassembler error>");
		return 0;
	}
	*obuff++ = '$';
	puthex(obuff, op);
	return 0;
}

static
OP_sI(int bytemode, char *obuff)
{
	int             op;

	switch (bytemode) {
	case b_mode:
		op = (schar_t) ADVANCE();
		break;
	case v_mode:
		if (dflag)
			op = get32();
		else
			op = (short) get16();
		break;
	case w_mode:
		op = (short) get16();
		break;
	default:
		strcpy(obuff, "<internal disassembler error>");
		return 0;
	}
	*obuff++ = '$';
	puthex(obuff, op);
	return 0;
}

static
OP_J(int bytemode, char *obuff)
{
	int             disp;
	int             mask = -1;

	switch (bytemode) {
	case b_mode:
		disp = (schar_t) ADVANCE();
		break;
	case v_mode:
		if (dflag)
			disp = get32();
		else {
			disp = (short) get16();
			/*
			 * for some reason, a data16 prefix on a jump
			 * instruction means that the pc is masked to 16 bits
			 * after the displacement is added!
			 */
			mask = 0xffff;
		}
		break;
	default:
		strcpy(obuff, "<internal disassembelr error>");
		return 0;
	}
	puthex(obuff,(BEGIN() + OFFSET() + disp) & mask);
	return 0;
}


static
OP_DIR(int size, char *obuff)
{
	int             seg, offset;

	switch (size) {
	case lptr:
		if (aflag) {
			offset = get32();
			seg = get16();
		} else {
			offset = get16();
			seg = get16();
		}
		obuff = puthex(obuff, seg);
		*obuff++ = ',';
		obuff = puthex(obuff, offset);
		break;
	case v_mode:
		if (aflag)
			offset = get32();
		else
			offset = (short) get16();
		obuff = puthex(obuff, BEGIN() + OFFSET() + offset);
		break;
	default:
		strcpy(obuff, "<internal disassembler error>");
		break;
	}
	return 0;
}

static
OP_OFF(char *obuff)
{
	int             off;
	if (aflag)
		off = get32();
	else
		off = get16();
	puthex(obuff, off);
	return 0;
}

static
do_op(int opid, int fl, char *opbuf)
{
	char	*op, obuff[100];
	*(op = obuff) = 0;

	switch (opid) {
	case 'E':
		*op++ = '*';
		/*FALLTHROUGH*/
	case 'e':
		OP_E(fl, op);
		break;
	case 'g':
		OP_G(fl, op);
		break;
	case 'r':
		strcpy(op, regname(rm, fl == d_mode));
		break;
	case 'i':
		OP_I(fl, op);
		break;
	case 'I':
		OP_sI(fl, op);
		break;
	case 'j':
		OP_J(fl, op);
		break;
	case '1':
		*op++ = '1';
		*op = '\0';
		break;
	case 'c':
		ADVANCE();		/* skip mod/rm */
		if (disasm_mode == 0)
			*op++ = '%';
		strcpy(op, "cr");
		ltoa(reg, op+2, 10);
		break;
	case 'd':
		ADVANCE();		/* skip mod/rm */
		if (disasm_mode == 0)
			*op++ = '%';
		strcpy(op, "db");
		ltoa(reg, op+2, 10);
		break;
	case 't':
		ADVANCE();		/* skip mod/rm */
		if (disasm_mode == 0)
			*op++ = '%';
		strcpy(op, "tr");
		ltoa(reg, op+2, 10);
		break;
	case 'R':
		OP_REG(fl, op);
		break;
	case 'S':
		strcpy(op,names_seg[reg] + disasm_mode);
		break;
	case 'D':
		OP_DIR(fl, op);
		break;
	case 'O':
		OP_OFF(op);
		break;
	case 'm':
		strcpy(op, segname(es_reg));
		op += strlen(op);
		*op++ = ':';
		*op++ = '(';
		strcpy(op, regname(rnum_esi, aflag));
		*op++ = ')';
		*op = '\0';
		break;
	case 'M':
		strcpy(op, segname(ds_reg));
		op += strlen(op);
		*op++ = ':';
		*op++ = '(';
		strcpy(op, regname(rnum_edi, aflag));
		*op++ = ')';
		*op = '\0';
		break;
	case 'f':
		strcpy(op, "%st" + disasm_mode);
		break;
	case 'F':
		if (disasm_mode == 0)
			*op++ = '%';
		strcpy(op, "st(");
		ltoa(rm, op+3, 10);
		op += strlen(op);
		*op++ = ')'; *op = 0;
		break;
	}
	strcat(opbuf, obuff);
	return 0;
}

static
dofloat(char *op, char *op1out, char *op2out, char *op3out)
{
	struct dis386  *dp;
	unsigned char   floatop;

	floatop = PREVIOUS();

	if (mod != 3) {
		putop(float_mem[(floatop - 0xd8) * 8 + reg], op);
		do_op('e', v_mode, op1out);
		return;
	}
	ADVANCE();

	dp = &float_reg[floatop - 0xd8][reg];
	if (dp->name == NULL) {
		putop(fgrps[dp->bytemode1][rm], op);
		/* instruction fnstsw is only one with strange arg */
		if (floatop == 0xdf && CURRENT() == 0xe0)
			strcpy(op1out, regname(rnum_eax, 1));
	} else {
		putop(dp->name, op);
		if (dp->op1)
			do_op(dp->op1, dp->bytemode1, op1out);
		if (dp->op2)
			do_op(dp->op2, dp->bytemode2, op2out);
	}
}

static
ckprefix()
{
	prefixes = 0;
	while (1) {
		switch (CURRENT()) {
		case 0xf3:
			prefixes |= PREFIX_REPZ;
			break;
		case 0xf2:
			prefixes |= PREFIX_REPNZ;
			break;
		case 0xf0:
			prefixes |= PREFIX_LOCK;
			break;
		case 0x2e:
			prefixes |= PREFIX_CS;
			break;
		case 0x36:
			prefixes |= PREFIX_SS;
			break;
		case 0x3e:
			prefixes |= PREFIX_DS;
			break;
		case 0x26:
			prefixes |= PREFIX_ES;
			break;
		case 0x64:
			prefixes |= PREFIX_FS;
			break;
		case 0x65:
			prefixes |= PREFIX_GS;
			break;
		case 0x66:
			prefixes |= PREFIX_DATA;
			break;
		case 0x67:
			prefixes |= PREFIX_ADR;
			break;
		case 0x9b:
			prefixes |= PREFIX_FWAIT;
			break;
		default:
			return;
		}
		ADVANCE();
	}
}



static char     op1out[100],	/* oprnd buffers */
                op2out[100],
                op3out[100],
                obuf[100];
static char    *obufp;		/* output pointer */


i386dis(pc, ibuf, outbuf, is32)
	int             pc;
	char             *ibuf;
char           *outbuf;
{
	struct dis386  *dp;
	int             i;
	int             enter_instruction;
	char           *first, *second, *third;
	int             needcomma;

	obuf[0] = 0;
	op1out[0] = 0;
	op2out[0] = 0;
	op3out[0] = 0;
	INIT(ibuf, pc);
	ckprefix();
	if (CURRENT() == 0xc8)
		enter_instruction = 1;
	else
		enter_instruction = 0;

	obufp = obuf;

	if (prefixes & PREFIX_REPZ)
		strcat(obufp, "repz ");
	if (prefixes & PREFIX_REPNZ)
		strcat(obufp, "repnz ");
	if (prefixes & PREFIX_LOCK)
		strcat(obufp,"lock ");

	if ((prefixes & PREFIX_FWAIT)
	    && ((CURRENT() < 0xd8) || (CURRENT() > 0xdf))) {
		/* fwait not followed by floating point instruction */
		strcat(obufp,"fwait");
		strcpy(outbuf, obuf);
		return 1;
	}
	/* these would be initialized to 0 if disassembling for 8086 or 286 */
	dflag = is32;
	aflag = is32;

	if (prefixes & PREFIX_DATA)
		dflag ^= 1;

	if (prefixes & PREFIX_ADR) {
		aflag ^= 1;
		strcat(obufp,"addr16 ");
	}
	if (CURRENT() == 0x0f) {
		ADVANCE();
		dp = &dis386_twobyte[dis_twoind[ADVANCE()]];
	} else {
		dp = &dis386[ADVANCE()];
	}
	mod = (CURRENT() >> 6) & 3;
	reg = (CURRENT() >> 3) & 7;
	rm = CURRENT() & 7;

	if (dp->name == NULL && dp->bytemode1 == FLOATCODE) {
		dofloat(obuf, op1out,op2out,op3out);
	} else {
		if (dp->name == NULL)
			dp = &grps[dp->bytemode1][reg];
		putop(dp->name, obuf);
		if (dp->op1)
			do_op(dp->op1, dp->bytemode1, op1out);

		if (dp->op2)
			do_op(dp->op2, dp->bytemode2, op2out);

		if (dp->op3)
			do_op(dp->op3, dp->bytemode3, op3out);
	}

	obufp = obuf + strlen(obuf);
	for (i = strlen(obuf); i < 6; i++)
		strcat(obufp, " ");
	strcat(obufp, " ");

	/*
	 * enter instruction is printed with operands in the same order as
	 * the intel book; everything else is printed in reverse order
	 */
	if (enter_instruction || disasm_mode) {
		first = op1out;
		second = op2out;
		third = op3out;
	} else {
		first = op3out;
		second = op2out;
		third = op1out;
	}
	needcomma = 0;
	if (*first) {
		strcat(obufp,first);
		needcomma = 1;
	}
	if (*second) {
		if (needcomma)
			strcat(obufp,",");
		strcat(obufp,second);
		needcomma = 1;
	}
	if (*third) {
		if (needcomma)
			strcat(obufp,",");
		strcat(obufp,third);
	}
	strcpy(outbuf, obuf);
	return OFFSET();
}



static
i386follow(pc, ibuf, is32, oc)
	int            pc;
	char          *ibuf;
	int            is32;
	OpCode        *oc;
{
	struct dis386  *dp;
	int             i;

	obuf[0] = 0;
	op1out[0] = 0;
	op2out[0] = 0;
	op3out[0] = 0;

	INIT(ibuf, pc);

	ckprefix();


	if ((prefixes & PREFIX_FWAIT)
	    && ((CURRENT() < 0xd8) || (CURRENT() > 0xdf))) {
		/* fwait not followed by floating point instruction */
		oc->flags = 0;
		oc->nfol = 1;
		oc->foll[0].flags = _OPR_ISREL;
		oc->foll[0].offset = OFFSET();
		return 0;
	}
	dflag = (is32!=0)^((prefixes&PREFIX_DATA)!=0);
	aflag = (is32!=0)^((prefixes&PREFIX_ADR)!=0);

	if (CURRENT() == 0x0f) {
		ADVANCE();
		dp = &dis386_twobyte[dis_twoind[ADVANCE()]];
	} else {
		dp = &dis386[ADVANCE()];
	}
	mod = (CURRENT() >> 6) & 3;
	reg = (CURRENT() >> 3) & 7;
	rm = CURRENT() & 7;

	oc->flags = 0;
	oc->nfol = 1;
	oc->foll[0].flags = _OPR_ISREL;

	if (dp->name == NULL && dp->bytemode1 == FLOATCODE) {
		dofloat(obuf, op1out,op2out,op3out);
		oc->foll[0].offset = OFFSET();
		return 0;
	}
	if (dp->name == NULL)
		dp = &grps[dp->bytemode1][reg];
	if (dp->name == NULL)
	putop(dp->name, obuf);
	if (dp->op1)
		do_op(dp->op1, dp->bytemode1, op1out);
	if (dp->op2)
		do_op(dp->op2, dp->bytemode2, op2out);
	if (dp->op3)
		do_op(dp->op3, dp->bytemode3, op3out);
/*-
 * jmp or call instruction.
 */
	oc->foll[0].offset = OFFSET();
	if (dp->op1 == 'j' || (dp->op1 == 'D' && dp->bytemode1==v_mode)) {
		oc->flags = _OP_ISJMP;
		if (strcmp(dp->name, "jmp") == 0) {
			oc->foll[0].offset = strtoul(op1out,0,0);
			return 0;
		} else {
			oc->nfol = 2;
			oc->foll[1].flags = _OPR_ISREL;
			oc->foll[1].offset= strtoul(op1out, 0, 0);
		}
		return 0;
	} else if (!strcmp(dp->name, "ret")  ||
	           !strcmp(dp->name, "iret") ||
	           !strcmp(dp->name, "lret") ||
	           !strcmp(dp->name, BAD)    ||
	           !strcmp(dp->name, "ljmp")) {
		oc->nfol = 0;
		return 0;
	}
	return 0;
}



analyze_opcode(uchar_t *op, OpCode *oc)
{

	return i386follow(0, op, 1, oc);
}



static int      lchar;
static int      start_pc;
static uchar_t *code_pointer;
static ulong_t  code_offset = 0;

static
INIT(char *ib, unsigned pc)
{
	start_pc = pc;
	code_pointer = ib;
	code_offset  = 0;
	lchar = -1;
	return 0;
}

static uchar_t
CURRENT()
{
	if (lchar == -1)
		lchar = code_pointer[code_offset];
	return lchar;
}

static          uchar_t
PREVIOUS()
{
	return code_pointer[code_offset-1];
}

static          uchar_t
ADVANCE()
{
	unsigned x = CURRENT();
	code_offset++;
	lchar = -1;
	return x;
}

static unsigned
OFFSET()
{
	return code_offset;
}

static unsigned
BEGIN()
{
	return start_pc;
}
